@@ -9,7 +9,7 @@
  * chet@ins.cwru.edu
  */
 
-/* Copyright (C) 1997-2016 Free Software Foundation, Inc.
+/* Copyright (C) 1997-2020 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -61,7 +61,7 @@
 		ae->next = new; \
 	} while (0)
 
-static char *array_to_string_internal __P((ARRAY_ELEMENT *, ARRAY_ELEMENT *, char *, int));
+static char *array_to_string_internal PARAMS((ARRAY_ELEMENT *, ARRAY_ELEMENT *, char *, int));
 
 static char *spacesep = " ";
 
@@ -398,16 +398,16 @@ ARRAY	*array;
  * Since arrays are sparse, unset array elements are not counted.
  */
 char *
-array_subrange (a, start, nelem, starsub, quoted)
+array_subrange (a, start, nelem, starsub, quoted, pflags)
 ARRAY	*a;
 arrayind_t	start, nelem;
-int	starsub, quoted;
+int	starsub, quoted, pflags;
 {
 	ARRAY		*a2;
 	ARRAY_ELEMENT	*h, *p;
 	arrayind_t	i;
-	char		*ifs, *sifs, *t;
-	int		slen;
+	char		*t;
+	WORD_LIST	*wl;
 
 	p = a ? array_head (a) : 0;
 	if (p == 0 || array_empty (a) || start > array_max_index(a))
@@ -432,32 +432,12 @@ int	starsub, quoted;
 
 	a2 = array_slice(a, h, p);
 
-	if (quoted & (Q_DOUBLE_QUOTES|Q_HERE_DOCUMENT))
-		array_quote(a2);
-	else
-		array_quote_escapes(a2);
-
-	if (starsub && (quoted & (Q_DOUBLE_QUOTES|Q_HERE_DOCUMENT))) {
-		/* ${array[*]} */
-		array_remove_quoted_nulls (a2);
-		sifs = ifs_firstchar ((int *)NULL);
-		t = array_to_string (a2, sifs, 0);
-		free (sifs);
-	} else if (quoted & (Q_DOUBLE_QUOTES|Q_HERE_DOCUMENT)) {
-		/* ${array[@]} */
-		sifs = ifs_firstchar (&slen);
-		ifs = getifs ();
-		if (ifs == 0 || *ifs == 0) {
-			if (slen < 2)
-				sifs = xrealloc(sifs, 2);
-			sifs[0] = ' ';
-			sifs[1] = '\0';
-		}
-		t = array_to_string (a2, sifs, 0);
-		free (sifs);
-	} else
-		t = array_to_string (a2, " ", 0);
+	wl = array_to_word_list(a2);
 	array_dispose(a2);
+	if (wl == 0)
+		return (char *)NULL;
+	t = string_list_pos_params(starsub ? '*' : '@', wl, quoted, pflags);	/* XXX */
+	dispose_words(wl);
 
 	return t;
 }
@@ -468,50 +448,29 @@ ARRAY	*a;
 char	*pat, *rep;
 int	mflags;
 {
-	ARRAY		*a2;
-	ARRAY_ELEMENT	*e;
-	char	*t, *sifs, *ifs;
-	int	slen;
+	char	*t;
+	int	pchar, qflags, pflags;
+	WORD_LIST	*wl, *save;
 
 	if (a == 0 || array_head(a) == 0 || array_empty(a))
 		return ((char *)NULL);
 
-	a2 = array_copy(a);
-	for (e = element_forw(a2->head); e != a2->head; e = element_forw(e)) {
-		t = pat_subst(element_value(e), pat, rep, mflags);
-		FREE(element_value(e));
-		e->value = t;
+	wl = array_to_word_list(a);
+	if (wl == 0)
+		return (char *)NULL;
+
+	for (save = wl; wl; wl = wl->next) {
+		t = pat_subst (wl->word->word, pat, rep, mflags);
+		FREE (wl->word->word);
+		wl->word->word = t;
 	}
 
-	if (mflags & MATCH_QUOTED)
-		array_quote(a2);
-	else
-		array_quote_escapes(a2);
+	pchar = (mflags & MATCH_STARSUB) == MATCH_STARSUB ? '*' : '@';
+	qflags = (mflags & MATCH_QUOTED) == MATCH_QUOTED ? Q_DOUBLE_QUOTES : 0;
+	pflags = (mflags & MATCH_ASSIGNRHS) ? PF_ASSIGNRHS : 0;
 
-	if (mflags & MATCH_STARSUB) {
-		array_remove_quoted_nulls (a2);
-		if ((mflags & MATCH_QUOTED) == 0 && ifs_is_null)
-			sifs = spacesep;
-		else
-			sifs = ifs_firstchar((int *)NULL);
-		t = array_to_string (a2, sifs, 0);
-		if (sifs != spacesep)
-			free(sifs);
-	} else if (mflags & MATCH_QUOTED) {
-		/* ${array[@]} */
-		sifs = ifs_firstchar (&slen);
-		ifs = getifs ();
-		if (ifs == 0 || *ifs == 0) {
-			if (slen < 2)
-				sifs = xrealloc (sifs, 2);
-			sifs[0] = ' ';
-			sifs[1] = '\0';
-		}
-		t = array_to_string (a2, sifs, 0);
-		free(sifs);
-	} else
-		t = array_to_string (a2, " ", 0);
-	array_dispose (a2);
+	t = string_list_pos_params (pchar, save, qflags, pflags);
+	dispose_words(save);
 
 	return t;
 }
@@ -523,53 +482,33 @@ char	*pat;
 int	modop;
 int	mflags;
 {
-	ARRAY		*a2;
-	ARRAY_ELEMENT	*e;
-	char	*t, *sifs, *ifs;
-	int	slen;
+	char	*t;
+	int	pchar, qflags, pflags;
+	WORD_LIST	*wl, *save;
 
 	if (a == 0 || array_head(a) == 0 || array_empty(a))
 		return ((char *)NULL);
 
-	a2 = array_copy(a);
-	for (e = element_forw(a2->head); e != a2->head; e = element_forw(e)) {
-		t = sh_modcase(element_value(e), pat, modop);
-		FREE(element_value(e));
-		e->value = t;
+	wl = array_to_word_list(a);
+	if (wl == 0)
+		return ((char *)NULL);
+
+	for (save = wl; wl; wl = wl->next) {
+		t = sh_modcase(wl->word->word, pat, modop);
+		FREE(wl->word->word);
+		wl->word->word = t;
 	}
 
-	if (mflags & MATCH_QUOTED)
-		array_quote(a2);
-	else
-		array_quote_escapes(a2);
+	pchar = (mflags & MATCH_STARSUB) == MATCH_STARSUB ? '*' : '@';
+	qflags = (mflags & MATCH_QUOTED) == MATCH_QUOTED ? Q_DOUBLE_QUOTES : 0;
+	pflags = (mflags & MATCH_ASSIGNRHS) ? PF_ASSIGNRHS : 0;
 
-	if (mflags & MATCH_STARSUB) {
-		array_remove_quoted_nulls (a2);
-		if ((mflags & MATCH_QUOTED) == 0 && ifs_is_null)
-			sifs = spacesep;
-		else
-			sifs = ifs_firstchar((int *)NULL);
-		t = array_to_string (a2, sifs, 0);
-		if (sifs != spacesep)
-			free(sifs);
-	} else if (mflags & MATCH_QUOTED) {
-		/* ${array[@]} */
-		sifs = ifs_firstchar (&slen);
-		ifs = getifs ();
-		if (ifs == 0 || *ifs == 0) {
-			if (slen < 2)
-				sifs = xrealloc (sifs, 2);
-			sifs[0] = ' ';
-			sifs[1] = '\0';
-		}
-		t = array_to_string (a2, sifs, 0);
-		free(sifs);
-	} else
-		t = array_to_string (a2, " ", 0);
-	array_dispose (a2);
+	t = string_list_pos_params (pchar, save, qflags, pflags);
+	dispose_words(save);
 
 	return t;
 }
+
 /*
  * Allocate and return a new array element with index INDEX and value
  * VALUE.
@@ -935,6 +874,60 @@ int	quoted;
 	return(result);
 }
 
+char *
+array_to_kvpair (a, quoted)
+ARRAY	*a;
+int	quoted;
+{
+	char	*result, *valstr, *is;
+	char	indstr[INT_STRLEN_BOUND(intmax_t) + 1];
+	ARRAY_ELEMENT *ae;
+	int	rsize, rlen, elen;
+
+	if (a == 0 || array_empty (a))
+		return((char *)NULL);
+
+	result = (char *)xmalloc (rsize = 128);
+	result[rlen = 0] = '\0';
+
+	for (ae = element_forw(a->head); ae != a->head; ae = element_forw(ae)) {
+		is = inttostr (element_index(ae), indstr, sizeof(indstr));
+		valstr = element_value (ae) ?
+				(ansic_shouldquote (element_value (ae)) ?
+				   ansic_quote (element_value(ae), 0, (int *)0) :
+				   sh_double_quote (element_value (ae)))
+					    : (char *)NULL;
+		elen = STRLEN (is) + 8 + STRLEN (valstr);
+		RESIZE_MALLOCED_BUFFER (result, rlen, (elen + 1), rsize, rsize);
+
+		strcpy (result + rlen, is);
+		rlen += STRLEN (is);
+		result[rlen++] = ' ';
+		if (valstr) {
+			strcpy (result + rlen, valstr);
+			rlen += STRLEN (valstr);
+		} else {
+			strcpy (result + rlen, "\"\"");
+			rlen += 2;
+		}
+
+		if (element_forw(ae) != a->head)
+		  result[rlen++] = ' ';
+
+		FREE (valstr);
+	}
+	RESIZE_MALLOCED_BUFFER (result, rlen, 1, rsize, 8);
+	result[rlen] = '\0';
+
+	if (quoted) {
+		/* This is not as efficient as it could be... */
+		valstr = sh_single_quote (result);
+		free (result);
+		result = valstr;
+	}
+	return(result);
+}
+
 char *
 array_to_assign (a, quoted)
 ARRAY	*a;
